const std = @import("std");
const clap = @import("clap");
const md4c = @import("md4c_h.zig");
const ansi = @import("ansi-term");
const cstdio = @cImport({
    @cInclude("stdio.h");
});

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();

    const params = comptime clap.parseParamsComptime(
        \\-h, --help display this help and exit
        \\<FILE> file to read from, if empty or "-" use stdin.
        \\
    );

    const parsers = comptime .{
        .FILE = clap.parsers.string,
    };

    var diag = clap.Diagnostic{};
    var res = clap.parse(clap.Help, &params, parsers, .{
        .diagnostic = &diag,
        .allocator = gpa.allocator(),
    }) catch |err| {
        diag.report(std.io.getStdErr().writer(), err) catch {};
        return err;
    };
    defer res.deinit();

    var reader: std.io.AnyReader = undefined;
    if (res.positionals.len < 1) {
        reader = std.io.getStdIn().reader().any();
    } else {
        const filename = res.positionals[0];
        const cwd = std.fs.cwd();
        reader = (cwd.openFile(filename, .{}) catch errorexit(13, "failed to open file D:\n")).reader().any();
    }

    const stdout_file = std.io.getStdOut().writer();
    var bw = std.io.bufferedWriter(stdout_file);
    const stdout = bw.writer();

    var state = State{
        .writer = stdout.any(),
    };

    const data = reader.readAllAlloc(gpa.allocator(), std.math.maxInt(usize)) catch errorexit(5, "failed to read from input stream D:");

    _ = md4c.md_parse(@ptrCast(data), @intCast(data.len), &parser, @ptrCast(&state));

    gpa.allocator().free(data);

    try stdout.writeByte('\n');

    try bw.flush();
}

const State = struct {
    writer: std.io.AnyWriter,
    last_style: ansi.style.Style = .{},
    nested_bold_count: u16 = 0,
    nested_italic_count: u16 = 0,
    nested_strike_count: u16 = 0,
    nested_underline_count: u16 = 0,

    fn update_style(self: *State, new: ansi.style.Style) void {
        ansi.format.updateStyle(self.writer, new, self.last_style) catch errorexit(1, "failed to output ansi style D:");
        self.last_style = new;
    }

    fn update_bold(self: *State) void {
        var temp_style = self.last_style;
        temp_style.font_style.bold = self.nested_bold_count > 0;
        self.update_style(temp_style);
    }

    fn update_italic(self: *State) void {
        var temp_style = self.last_style;
        temp_style.font_style.italic = self.nested_italic_count > 0;
        self.update_style(temp_style);
    }

    fn update_strike(self: *State) void {
        var temp_style = self.last_style;
        temp_style.font_style.crossedout = self.nested_strike_count > 0;
        self.update_style(temp_style);
    }

    fn update_underline(self: *State) void {
        var temp_style = self.last_style;
        temp_style.font_style.underline = self.nested_underline_count > 0;
        self.update_style(temp_style);
    }
};

fn enter_block(blocktype: md4c.MD_BLOCKTYPE, detail: ?*anyopaque, userdata: ?*anyopaque) callconv(.C) c_int {
    //std.debug.print("enter {d}\n", .{blocktype});
    const state: *State = @ptrCast(@alignCast(userdata));

    if (blocktype == md4c.MD_BLOCK_H) {
        just_write(state.writer, "\n");
        const h_detail: *md4c.MD_BLOCK_H_DETAIL = @ptrCast(@alignCast(detail));
        var temp_style = state.last_style;
        temp_style.font_style.dim = true;
        temp_style.font_style.bold = true;
        state.update_style(temp_style);
        for (0..h_detail.level) |_| {
            just_write(state.writer, "#");
        }
        temp_style.font_style.dim = false;
        temp_style.font_style.bold = state.nested_bold_count > 0;
        state.update_style(temp_style);
        just_write(state.writer, " ");
    }
    if (blocktype == md4c.MD_BLOCK_LI) {
        const li_detail: *md4c.MD_BLOCK_LI_DETAIL = @ptrCast(@alignCast(detail));
        var temp_style = state.last_style;
        if (li_detail.is_task == 0) temp_style.font_style.dim = true;
        temp_style.font_style.bold = true;
        state.update_style(temp_style);
        if (li_detail.is_task == 0) {
            just_write(state.writer, "• ");
        } else {
            if (li_detail.task_mark == 'x' or li_detail.task_mark == 'X') {
                just_write(state.writer, "☑ ");
            } else {
                just_write(state.writer, "☐ ");
            }
        }
        temp_style.font_style.dim = false;
        temp_style.font_style.bold = state.nested_bold_count > 0;
        state.update_style(temp_style);
    }
    if (blocktype == md4c.MD_BLOCK_UL or blocktype == md4c.MD_BLOCK_OL) {
        just_write(state.writer, "\n");
    }
    return 0; // TODO what does the return value represent ????
}

fn leave_block(blocktype: md4c.MD_BLOCKTYPE, detail: ?*anyopaque, userdata: ?*anyopaque) callconv(.C) c_int {
    //std.debug.print("leave {d}\n", .{blocktype});
    const state: *State = @ptrCast(@alignCast(userdata));

    if (blocktype == md4c.MD_BLOCK_H) {
        just_write(state.writer, "\n\n");
    }
    if (blocktype == md4c.MD_BLOCK_P or blocktype == md4c.MD_BLOCK_LI) {
        just_write(state.writer, "\n");
    }
    if (blocktype == md4c.MD_BLOCK_HR) {
        var winsize: std.c.winsize = undefined;
        var temp_style = state.last_style;
        temp_style.font_style.dim = true;
        temp_style.font_style.bold = true;
        state.update_style(temp_style);
        const term_name_c = cstdio.ctermid(null);
        const term_name = std.mem.sliceTo(term_name_c, 0);
        const term_fd = std.posix.open(term_name, .{}, 0) catch errorexit(5, "failed to open terminal D:\n");
        if (std.c.ioctl(term_fd, std.c.T.IOCGWINSZ, &winsize) != -1) {
            for (0..winsize.ws_col) |_| {
                just_write(state.writer, "-");
            }
            just_write(state.writer, "\n");
        } else {
            just_write(state.writer, "---\n");
        }
        temp_style.font_style.dim = false;
        temp_style.font_style.bold = state.nested_bold_count > 0;
        state.update_style(temp_style);
    }
    if (blocktype == md4c.MD_BLOCK_UL or blocktype == md4c.MD_BLOCK_OL) {
        just_write(state.writer, "\n");
    }
    _ = detail;
    return 0; // TODO what does the return value represent ????
}

fn enter_span(spantype: md4c.MD_SPANTYPE, detail: ?*anyopaque, userdata: ?*anyopaque) callconv(.C) c_int {
    const state: *State = @ptrCast(@alignCast(userdata));
    if (spantype == md4c.MD_SPAN_STRONG) {
        state.nested_bold_count +|= 1;
        state.update_bold();
    }
    if (spantype == md4c.MD_SPAN_EM) {
        state.nested_italic_count +|= 1;
        state.update_italic();
    }
    if (spantype == md4c.MD_SPAN_DEL) {
        state.nested_strike_count +|= 1;
        state.update_strike();
    }
    if (spantype == md4c.MD_SPAN_U) {
        state.nested_underline_count +|= 1;
        state.update_underline();
    }
    if (spantype == md4c.MD_SPAN_A) {
        var temp_style = state.last_style;
        temp_style.foreground = .Blue;
        temp_style.font_style.underline = true;
        state.update_style(temp_style);
        just_write(state.writer, "\x1b]8;;");
        const detail_a: *md4c.MD_SPAN_A_DETAIL = @ptrCast(@alignCast(detail));
        const link_c = detail_a.href.text;
        const link = link_c[0..detail_a.href.size];
        just_write(state.writer, link);
        just_write(state.writer, "\x07");
    }
    // TODO emit ansi to the writer
    return 0; // TODO what does the return value represent ????
}

fn leave_span(spantype: md4c.MD_SPANTYPE, detail: ?*anyopaque, userdata: ?*anyopaque) callconv(.C) c_int {
    _ = detail;
    const state: *State = @ptrCast(@alignCast(userdata));
    if (spantype == md4c.MD_SPAN_STRONG) {
        state.nested_bold_count -|= 1;
        state.update_bold();
    }
    if (spantype == md4c.MD_SPAN_EM) {
        state.nested_italic_count -|= 1;
        state.update_italic();
    }
    if (spantype == md4c.MD_SPAN_DEL) {
        state.nested_strike_count -|= 1;
        state.update_strike();
    }
    if (spantype == md4c.MD_SPAN_U) {
        state.nested_underline_count -|= 1;
        state.update_underline();
    }
    if (spantype == md4c.MD_SPAN_A) {
        just_write(state.writer, "\x1b]8;;\x07");
        var temp_style = state.last_style;
        temp_style.foreground = .Default;
        state.update_style(temp_style);
        state.update_underline();
    }
    // TODO emit ansi to the writer
    return 0; // TODO what does the return value represent ????
}

fn on_text(texttype: md4c.MD_TEXTTYPE, text: [*c]const md4c.MD_CHAR, size: md4c.MD_SIZE, userdata: ?*anyopaque) callconv(.C) c_int {
    const state: *State = @ptrCast(@alignCast(userdata));

    const text_data: []const u8 = @ptrCast(text[0..size]);

    if (texttype == md4c.MD_TEXT_NORMAL) {
        just_write(state.writer, text_data);
    }
    if (texttype == md4c.MD_TEXT_NULLCHAR) {
        just_write(state.writer, "�");
    }
    if (texttype == md4c.MD_TEXT_BR) {
        just_write(state.writer, "\n");
    }
    if (texttype == md4c.MD_TEXT_CODE) {
        var temp_style = state.last_style;
        temp_style.font_style.dim = true;
        state.update_style(temp_style);
        just_write(state.writer, text_data);
        temp_style.font_style.dim = false;
        state.update_style(temp_style);
    }
    return 0; // TODO what does the return value represent ????
}

const parser: md4c.MD_PARSER = .{
    .abi_version = @intCast(0),
    .flags = md4c.MD_FLAG_UNDERLINE | md4c.MD_FLAG_STRIKETHROUGH | md4c.MD_FLAG_COLLAPSEWHITESPACE | md4c.MD_FLAG_TASKLISTS | md4c.MD_FLAG_UNDERLINE,

    .enter_block = &enter_block,
    .leave_block = &leave_block,
    .enter_span = &enter_span,
    .leave_span = &leave_span,
    .text = &on_text,

    .debug_log = null,
    .syntax = null,
};

fn errorexit(code: u8, comptime msg: []const u8) noreturn {
    std.debug.print(msg, .{});
    std.process.exit(code);
}

fn just_write(writer: std.io.AnyWriter, data: []const u8) void {
    writer.writeAll(data) catch errorexit(5, "failed to write to output D:\n");
}
